/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "hdrs.h"
#include "pkt.h"

#include <linux/inetdevice.h>
#include <net/route.h>
#include <net/ip6_route.h>


int pkt_verdict(struct sk_buff *skb, int verdict)
{
	bool policy_routing = false;
	struct nf_queue_entry *nf_data = NULL;
	struct sk_buff *next = NULL;

	if (skb->kdata.type == NET_DEV_PACKET) {
		TRACE_DBG("%s: skb %p describing net device freed.\n", __func__, skb);
		kfree_skb(skb);
		return 0;
	}
	next = skb->kdata.next;
	if (next) {
		/* first free reasm packet */
		TRACE_DBG("%s: reasembled skb %p will be freed.\n", __func__, skb);
		kfree_skb(skb);
		skb = next;
	}
	while (skb) {
		next = skb->kdata.next;
		if (skb->kdata.type != NETWORK_PACKET) {
			TRACE(LOG_WARN, "%s: skb %p with unknown type 0x%08X freed.", __func__, skb, skb->kdata.type);
			kfree_skb(skb);
			skb = next;
			continue;
		}
		nf_data = skb->kdata.nf_data;
		if (!nf_data) {
			TRACE(LOG_WARN, "%s: skb %p with invalid nf_data pointer freed.", __func__, skb);
			kfree_skb(skb);
			skb = next;
			continue;
		}
		/* fix reasm skb */
		if (nf_data->skb != skb) {
			nf_data->skb = skb;
		}
		policy_routing = skb->kdata.policy_routing;
		
		/* clear only portion affected by our code */
		memset(&skb->kdata, 0, sizeof(struct kipf_skb_data));

		TRACE_DBG("%s: skb %p %s on '%s' %s\n", __func__, skb, ((nf_data->hook == NF_IP_PRE_ROUTING) ? ("received") : ("sent")),
			((skb->dev) ? (skb->dev->name) : ("???")), ((verdict == NF_DROP) ? ("dropped") : (verdict == NF_ACCEPT) ? ("accepted") :
			(verdict == NF_STOLEN) ? ("stolen") : (verdict == NF_QUEUE) ? ("queued") : (verdict == NF_REPEAT) ? ("repeated") : ("???")));

		if (policy_routing) {
			if (verdict == NF_ACCEPT) {
				int ret = 0;

				verdict = NF_STOLEN;
				nf_reinject(nf_data, verdict);
				TRACE_DBG("%s: policy routing applied on skb %p\n", __func__, skb);

				/* we don't want to see indicated packet again */
				skb->mark = PKT_SOCKET_TAG;
				
				/* we can't simply call nf_reinject, because snort have to inspect this skb */
				if (skb->protocol == htons(ETH_P_IPV6)) {
					if ((ret = NF_HOOK_COND(PF_INET6, NF_INET_POST_ROUTING, skb, NULL, skb->dev, ip6_finish_output, 1))) {
						TRACE(LOG_WARN, "%s: ip6_finish_output(): %d, skb %p.", __func__, ret, skb);
					}
				}
				else {
					if ((ret = NF_HOOK_COND(PF_INET, NF_INET_POST_ROUTING, skb, NULL, skb->dev, ip_finish_output, 1))) {
						TRACE(LOG_WARN, "%s: ip_finish_output(): %d, skb %p.", __func__, ret, skb);
					}
				}
			} else {
				if (verdict != NF_DROP) {
					/* we don't allow verdict other then NF_DROP for policy routed packets */
					TRACE(LOG_WARN, "%s: skb %p dropped due to invalid verdict on policy routed packet.", __func__, skb);
					verdict = NF_DROP;
				}
				nf_reinject(nf_data, verdict);
			}
		} else {
			nf_reinject(nf_data, verdict);
		}
		/* go to next fragment, if any */
		skb = next;
	}
	return 0;
}

static int32_t pkt_policy_routing(struct sk_buff *skb, struct kipf_route_info *ri, bool rcv)
{
	struct iphdr *iph = ip_hdr(skb);

	struct sk_buff *frag;
	__be32 rt_ip;
	unsigned short hh_len; 

	/* policy routing by IP */
	if (ri->next_hop) {
		rt_ip = ri->next_hop;
	}
	/* policy routing by ifindex */
	else if (ri->ifindex) {
		int found = 0;
		struct in_device *in_dev = NULL;
		struct in_ifaddr *ifa = NULL;
		struct in_ifaddr **ifap = NULL;
		struct net_device *dev = netdev_find_by_ifindex(ri->ifindex);

		if (!IS_PTR_VALID(dev)) {
			TRACE(LOG_WARN, "%s: can't find net device with ifindex %d.", __func__, ri->ifindex);
			return -EINVAL;
		}

		if (dev->type != ARPHRD_PPP) {
			TRACE(LOG_WARN, "%s: policy routing is not allowed on non ppp device '%s'.", __func__, dev->name);
			return -EINVAL;
		}

		/* in real just return (struct in_device *)dev->ip_ptr */
		if (!(in_dev = __in_dev_get_rtnl(dev))) {
			TRACE(LOG_WARN, "%s: __in_dev_get_rtnl(): can't get in device for '%s'.", __func__, dev->name);
			return -EINVAL;
		}

		/* ToDo: just take first destination IP ? */
		for (ifap = &in_dev->ifa_list; (ifa = *ifap) != NULL; ifap = &ifa->ifa_next) {
			if (!strcmp(dev->name, ifa->ifa_label)) {
				rt_ip = ifa->ifa_address;
				found = 1;
				break;
			}
		}

		if (!found) {
			TRACE(LOG_WARN, "%s: can't get destination ip for '%s'.", __func__, dev->name);
			return -EINVAL;
		}
	}
	/* no policy routing */
	else {
		return -EINVAL;
	}

	/*
	 * see ip_route_me_harder(). Also skb->dst can be NULL:
 	 * - if called from pkt_send_out()
	 * - if called on received packet, skb->dst is normally filled later in ip_rcv_finish()
  	 */
	TRACE_DBG("policy_routing - current dst:%p, dev: %p\n", skb_dst(skb), skb->dev);

	if (rcv) {
		int err;

		skb_dst_drop(skb);
		err = ip_route_input_noref(skb, rt_ip, iph->saddr, iph->tos, skb->dev); // TODO: noref?
		if (err) {
			WARN(1, "ip_route_input_noref failed\n");
			return -EINVAL;
		}
	}
	else {
		struct rtable *rt;
		int err;
		struct flowi4 fl4 = {
			.daddr = rt_ip,
			.saddr = iph->saddr,
			.flowi4_tos = iph->tos,
			.flowi4_flags = FLOWI_FLAG_ANYSRC,
			.flowi4_proto = iph->protocol
		};
		fl4.flowi4_oif = ri->ifindex;

		rt = ip_route_output_key(&init_net, &fl4);
		err = PTR_ERR(rt);
		if (IS_ERR_VALUE(err)) {
			WARN(1, "ip_route_output didn't return rt (%x) tr_ip: %x, saddr: %x, tos: %d\n", err, rt_ip, iph->saddr, iph->tos);
			return err;
		}
		skb_dst_drop(skb);
		skb_dst_set(skb, &rt->dst);
	}

	skb->dev = skb_dst(skb)->dev;
	skb->kct.next_hop = rt_ip;
	skb_walk_frags(skb, frag) {
		frag->kct.next_hop = rt_ip;
	}


	/* Change in oif may mean change in hh_len. (taken from ip_route_me_harder) */
	hh_len = skb->dev->hard_header_len;
	if (skb_headroom(skb) < hh_len) {
		int err = pskb_expand_head(skb, HH_DATA_ALIGN(hh_len - skb_headroom(skb)), 0, GFP_ATOMIC);
		WARN(err, "not enough room for hh_len after policy routing, reallocation failed!\n");
	}

	TRACE_DBG("%s: skb %p will be routed through %u.%u.%u.%u, dst %p\n", __func__, skb, IP_ADDRESS_STRING(rt_ip), skb_dst(skb));
	return 0;
}

static bool in6_addr_eq(const struct in6_addr l, const struct in6_addr r)
{
	return memcmp(&l, &r, sizeof(struct in6_addr)) == 0;
}

static int32_t pkt_policy_routing_ip6(struct sk_buff *skb, struct kipf_route_info *ri, bool rcv)
{
	struct ipv6hdr *iph = ipv6_hdr(skb);

	struct sk_buff *frag;
	struct	in6_addr rt_ip;
	unsigned short hh_len;
	struct dst_entry *dst;
	int err;
	
	/* policy routing by IP */
	if ( ! in6_addr_eq(ri->next_hop6, in6addr_any)) {
		rt_ip = ri->next_hop6;
	}
	/* policy routing by ifindex */
	else if (ri->ifindex) {
		struct inet6_dev *in_dev = NULL;
		struct inet6_ifaddr *ifa = NULL;
		struct net_device *dev = netdev_find_by_ifindex(ri->ifindex);

		if (!IS_PTR_VALID(dev)) {
			TRACE(LOG_WARN, "%s: can't find net device with ifindex %d.", __func__, ri->ifindex);
			return -EINVAL;
		}

		if (dev->type != ARPHRD_PPP) {
			TRACE(LOG_WARN, "%s: policy routing is not allowed on non ppp device '%s'.", __func__, dev->name);
			return -EINVAL;
		}

		/* in real just return (struct in_device *)dev->ip6_ptr */
		if (!(in_dev = __in6_dev_get(dev))) {
			TRACE(LOG_WARN, "%s: __in_dev_get_rtnl(): can't get in device for '%s'.", __func__, dev->name);
			return -EINVAL;
		}
			
		ifa = list_first_entry(&in_dev->addr_list, struct inet6_ifaddr, if_list);
		if ( ! ifa || in6_addr_eq(ifa->peer_addr, in6addr_any)) {
			TRACE(LOG_WARN, "%s: can't get destination ip for '%s'.", __func__, dev->name);
			return -EINVAL;
		}
		
		rt_ip = ifa->peer_addr;
	}
	/* no policy routing */
	else {
		return -EINVAL;
	}
	
	/*
	 * see ip_route_me_harder(). Also skb->dst can be NULL:
	 * - if called from pkt_send_out()
	 * - if called on received packet, skb->dst is normally filled later in ip_rcv_finish()
	 */
	TRACE_DBG("policy_routing_6 - current dst:%p, dev: %p\n", skb_dst(skb), skb->dev);
	
	if (rcv) {
		struct net *net = dev_net(skb->dev);
		int flags = RT6_LOOKUP_F_HAS_SADDR;
		struct flowi6 fl6 = {
			.flowi6_iif = skb->dev->ifindex,
			.daddr = rt_ip,
			.saddr = iph->saddr,
			.flowlabel = ip6_flowinfo(iph),
			.flowi6_mark = skb->mark,
			.flowi6_proto = iph->nexthdr,
		};

		dst = ip6_route_input_lookup(net, skb->dev, &fl6, flags);
	}
	else {
		struct flowi6 fl6 = {
			.daddr = rt_ip,
			.saddr = iph->saddr,
			.flowi6_flags = FLOWI_FLAG_ANYSRC,
			.flowi6_oif = ri->ifindex,
			.flowlabel = ip6_flowinfo(iph),
			.flowi6_mark = skb->mark,
			.flowi6_proto = iph->nexthdr,
		};

		dst = ip6_route_output(&init_net, NULL, &fl6);
	}
	
	err = PTR_ERR(dst);
	if (IS_ERR_VALUE(err)) {
		WARN(1, "%s didn't return dst (%x) daddr: %pI6c, saddr: %pI6c, priority: %d\n", rcv ? "ip6_route_input_lookup" : "ip6_route_output", err, &rt_ip, &iph->saddr, iph->priority);
		return err;
	}

	skb_dst_drop(skb);
	skb_dst_set(skb, dst);
	skb->dev = skb_dst(skb)->dev;
	skb->kct.next_hop6 = rt_ip;
	skb_walk_frags(skb, frag) {
		frag->kct.next_hop6 = rt_ip;
	}

	/* Change in oif may mean change in hh_len. (taken from ip_route_me_harder) */
	hh_len = skb->dev->hard_header_len;
	if (skb_headroom(skb) < hh_len) {
		int err = pskb_expand_head(skb, HH_DATA_ALIGN(hh_len - skb_headroom(skb)), 0, GFP_ATOMIC);
		WARN(err, "not enough room for hh_len after policy routing, reallocation failed!\n");
	}

	TRACE_DBG("%s: skb %p will be routed through %pI6c, dst %p\n", __func__, skb, &rt_ip, skb_dst(skb));
	return 0;
}

int32_t pkt_accept(struct kipf_packet *packet)
{
	struct sk_buff *skb = NULL;
	int ret = 0;
	int verdict = NF_DROP;
	void* packet_id;
	struct kipf_packet_u2k u2k;
	struct dst_entry *dst;

	if (get_user(packet_id, (void* __user *) &packet->id)) {
		TRACE(LOG_WARN, "%s: cannot get packet id, segmentation fault.", __func__);
		return -EINVAL;
	}

	if ((ret = copy_from_user(&u2k, (u_int8_t __user *) &packet->u2k, sizeof(struct kipf_packet_u2k)))) {
		TRACE(LOG_WARN, "%s: cannot get packet info, segmentation fault (%d).", __func__, ret);
		return -EINVAL;
	}

	do {
		if (!(skb = queue_remove(&g_pending_queue, (void*) packet_id, queue_cmp_by_id))) {
			TRACE(LOG_WARN, "%s: can't remove skb with id %p from pending queue.", __func__, (void*) packet_id);
			break;
		}

		skb->kct.used = 0;
		skb->kct.next_hop = 0;
		skb->kct.next_hop6 = in6addr_any;
		//TRACE_DBG("%s: skb %p removed from pending queue.\n", __func__, skb);

		if (skb->kdata.type != NETWORK_PACKET) {
			TRACE(LOG_WARN, "%s: skb %p dropped due to wrong type %08X", __func__, skb, skb->kdata.type);
			break;
		}

		/*	packet was modified in user space? */
		if (u2k.modif_len > 0) {
			struct iovec iov = { .iov_base = packet->data + sizeof(struct ethhdr), .iov_len = u2k.modif_len};

			if (u2k.modif_len > skb->len) {
				TRACE(LOG_WARN, "%s: skb %p dropped due to invalid length of modified part (%d).", __func__, skb, u2k.modif_len);
				break;
			}

			if (skb->protocol == htons(ETH_P_IP) && skb->len >= sizeof(struct ethhdr) + sizeof(struct iphdr)) {
				struct iphdr *ipv4 = ip_hdr(skb);

				if (skb->kdata.received) {
					skb->kct.used = 1;
					skb->kct.daddr = ipv4->daddr;
				}
				else if (u2k.route_info.next_hop == 0) {
					u2k.route_info.next_hop = ipv4->daddr;
					TRACE_DBG("%s: skb %p policy routing forced\n",__func__, skb);
				}
			}
			
			if (skb->protocol == htons(ETH_P_IPV6) && skb->len >= sizeof(struct ethhdr) + sizeof(struct ipv6hdr)) {
				struct ipv6hdr *ipv6 = ipv6_hdr(skb);

				if (skb->kdata.received) {
					skb->kct.used = 1;
					skb->kct.daddr6 = ipv6->daddr;
				}
				else if (in6_addr_eq(u2k.route_info.next_hop6, in6addr_any)) {
					u2k.route_info.next_hop6 = ipv6->daddr;
					TRACE_DBG("%s: skb %p policy routing forced\n",__func__, skb);
				}
			}

			if ((ret = skb_copy_datagram_from_iovec(skb, 0, &iov, 0, u2k.modif_len))) {
				TRACE(LOG_WARN, "%s: skb %p dropped due to error in skb_store_bits(): %d", __func__, skb, ret);
				break;
			}

			TRACE_DBG("%s: skb %p modified with %d byte(s).\n",__func__, skb, u2k.modif_len);

			if (skb_is_gso(skb) && skb->ip_summed == CHECKSUM_PARTIAL) {
				// since L4 checksum has to be partially summed, we do it manually in case of packet modification
				int offset = skb_checksum_start_offset(skb);
				BUG_ON(offset >= skb_headlen(skb));

				offset += skb->csum_offset;
				BUG_ON(offset + sizeof(__sum16) > skb_headlen(skb));

				if (skb->protocol == htons(ETH_P_IPV6)) {
					struct ipv6hdr *iph = ipv6_hdr(skb);
					*(__sum16 *)(skb->data + offset) = 0xFFFF - csum_ipv6_magic(&iph->saddr, &iph->daddr, ntohs(iph->payload_len), iph->nexthdr, 0);
				}
				else {
					struct iphdr *iph = ip_hdr(skb);
					*(__sum16 *)(skb->data + offset) = 0xFFFF - csum_tcpudp_magic(iph->saddr, iph->daddr, skb->len - sizeof(struct iphdr), iph->protocol, 0);
				}
			}
		}

		/* apply policy routing ? */
		if (u2k.route_info.ifindex || ((skb->protocol == htons(ETH_P_IP) && u2k.route_info.next_hop) || (skb->protocol == htons(ETH_P_IPV6) && ! in6_addr_eq(u2k.route_info.next_hop6, in6addr_any)))) {

			if (skb->protocol == htons(ETH_P_IPV6)) {
				if ((ret = pkt_policy_routing_ip6(skb, &u2k.route_info, skb->kdata.received))) {
					TRACE(LOG_WARN, "%s: skb %p dropped due to error in pkt_policy_routing_ip6(): %d", __func__, skb, ret);
					break;
				}
			}
			else {
				if ((ret = pkt_policy_routing(skb, &u2k.route_info, skb->kdata.received))) {
					TRACE(LOG_WARN, "%s: skb %p dropped due to error in pkt_policy_routing(): %d", __func__, skb, ret);
					break;
				}
			}

			/* modify nf_data to reflect policy routing - only for io_read_net_skb() now */
			{
				struct nf_queue_entry *nf_data = skb->kdata.nf_data;
				nf_data->hook = NF_INET_POST_ROUTING;
				/* indev/outdev is ref counted in /linux-2.6.38/net/netfilter/nf_queue.c!__nf_queue */
				if (nf_data->indev) {
					dev_put(nf_data->indev);
				}
				if (nf_data->outdev) {
					dev_put(nf_data->outdev);
				}
				nf_data->indev = NULL;
				nf_data->outdev = skb->dev;
				if (skb->protocol == htons(ETH_P_IPV6)) {
					nf_data->okfn = ip6_finish_output;
				}
				else {
					nf_data->okfn = ip_finish_output;
				}
				/* we don't want to see indicated packet again */
				skb->mark = PKT_SOCKET_TAG;

				/* increase ref count on new device */
				dev_hold(nf_data->outdev);
			}

			skb->kdata.policy_routing = true;

			/* received packets */
			if (skb->kdata.received) {

				WARN(1, "policy routing on inbound packet is deprecated!\n");

				/* we want to see received packets as sent */
				skb->kdata.received = false;

				if ((ret = queue_insert(&g_main_queue, skb, 1))) {
					TRACE(LOG_WARN, "%s: skb %p dropped due to disabled main queue.", __func__, skb);
					break;
				}

				TRACE_DBG("%s: skb %p re-inserted into main queue.\n", __func__, skb);
				io_dev_wakeup();

				/* avoid invocation of pkt_verdict() */
				skb = NULL;
				break;
			}
		}

		dst = skb_dst(skb);
		if (dst) {
			TRACE_DBG("%s: skb %p input %pf output %pf xfrm %p child %p path %p", __func__, skb, dst->input, dst->output, dst->xfrm, dst->child, dst->path);
		}

		verdict = NF_ACCEPT;

	} while(0);

	if (skb) {
		pkt_verdict(skb, verdict);
	}

	return ret;
}

int32_t pkt_drop(struct kipf_packet *packet)
{
	void* packet_id;
	struct sk_buff *skb = NULL;

	if (get_user(packet_id, (void* __user *) &packet->id)) {
		TRACE(LOG_WARN, "%s: cannot get packet id, segmentation fault.", __func__);
		return -EINVAL;
	}

	if (!(skb = queue_remove(&g_pending_queue, packet_id, queue_cmp_by_id))) {
		TRACE(LOG_WARN, "%s: can't remove skb with id %p from pending queue.", __func__, packet_id);
		return -EINVAL;
	}

	TRACE_DBG("%s: skb %p removed from pending queue.\n", __func__, skb);

	pkt_verdict(skb, NF_DROP);

	return 0;
}

int32_t pkt_complete(struct kipf_packet *packet)
{
	struct sk_buff *skb;
	int32_t ret = -EINVAL;
	void* packet_id;

	if (get_user(packet_id, (void* __user *) &packet->id)) {
		TRACE(LOG_WARN, "%s: cannot get packet id, segmentation fault.", __func__);
		return -EINVAL;
	}

	do {
		int frame_len;
		int bytes_up;
		int offset_in_skb;

		if (!(skb = queue_remove(&g_pending_queue, (void*) packet_id, queue_cmp_by_id))) {
			TRACE(LOG_WARN, "%s: can't remove skb with id %p from pending queue.", __func__, packet_id);
			ret = -EINVAL;
			break;
		}

		TRACE_DBG("%s: skb %p removed from pending queue.\n", __func__, skb);

		frame_len = sizeof(struct ethhdr) + skb->len;
		TRACE_DBG("%s: skb %p, frame_len %d\n", __func__, skb, frame_len);

		offset_in_skb = skb->kdata.copied_len - sizeof(struct ethhdr);
		bytes_up = (frame_len <= g_linipf_cfg.max_packet_data_size) ? (frame_len - skb->kdata.copied_len) : (g_linipf_cfg.max_packet_data_size - skb->kdata.copied_len);

		{
			struct iovec iov = { .iov_base = packet->data + skb->kdata.copied_len, .iov_len = bytes_up};

			if (skb_copy_datagram_iovec(skb, offset_in_skb, &iov, bytes_up)) {
				TRACE(LOG_WARN, "%s: skb_copy_bits(): can't copy %d bytes of data from skb %p due to error.", __func__, bytes_up, skb);
				ret = -EINVAL;
				break;
			}
		}

		skb->kdata.copied_len += bytes_up;

		//if (put_user(skb->kdata.copied_len, &packet->data_len)) {
		//	TRACE(LOG_WARN, "%s: skb %p dropped due to segmentation fault.", __func__, skb);
		//	ret = -EFAULT;
		//	break;
		//}

		TRACE_DBG("%s: skb %p completed - copied %d bytes from total %d bytes of data.\n", __func__, skb, bytes_up, skb->len);

		if ((ret = queue_insert(&g_pending_queue, skb, 0))) {
			TRACE(LOG_WARN, "%s: skb %p dropped due to disabled pending queue.", __func__, skb);
			break;
		}

		TRACE_DBG("%s: skb %p re-inserted into pending queue.\n", __func__, skb);

		ret = 0;

	} while(0);

	/* some error */
	if (ret && skb) {
		pkt_verdict(skb, NF_DROP);
	}

	return ret;
}

int32_t pkt_tx_ip(unsigned char *data, unsigned int data_len, struct kipf_route_info *route_info)
{
	struct sk_buff *skb = NULL;
	u_int32_t ret = 0;
	unsigned char ver, *new_skb_data;
	const uint extra_headroom = 15;
	struct iphdr ip_header;

	if (!data) {
		TRACE(LOG_WARN, "%s: invalid parameter.", __func__);
		return -EINVAL;
	}

	if (data_len < sizeof(struct ethhdr) + sizeof(struct iphdr)) {
		TRACE(LOG_WARN, "%s: packet length is too short: %d", __func__, data_len);
		return -EINVAL;
	}
	
	if ((ret = copy_from_user(&ip_header, data + sizeof(struct ethhdr), sizeof(struct iphdr)))) {
		TRACE(LOG_WARN, "%s: cannot get ip version, segmentation fault (%d).", __func__, ret);
		return -EINVAL;
	}

	ver = ip_header.version;

	if (ver != 4 && ver != 6) {
		TRACE(LOG_WARN, "%s: unsupported version of IP (%d).", __func__, ver);
		return -EINVAL;
	}
	
	if (ver == 6) {
		if (data_len < sizeof(struct ethhdr) + sizeof(struct ipv6hdr)) {
			TRACE(LOG_WARN, "%s: packet length is too short: %d", __func__, data_len);
			return -EINVAL;
		}
	}

	if (!(skb = alloc_skb(data_len + extra_headroom + NET_IP_ALIGN + NET_SKB_PAD, GFP_KERNEL))) {
		TRACE(LOG_WARN, "%s: dev_alloc_skb(): can't allocate %d bytes of memory.", __func__, data_len + NET_IP_ALIGN);
		return -ENOMEM;
	}

	TRACE_DBG("%s: skb %p allocated.\n", __func__, skb);

	skb->protocol = (ver == 4) ? htons(ETH_P_IP) : htons(ETH_P_IPV6);

	// increase headroom, so we must not move the data in case of additional headers (like PPPoE).
	// The value is poor's men guess, correct way is to determine dst first and use dst->dev->hard_header_len
	// (see __ip_append_data())
	skb_reserve(skb, sizeof(struct ethhdr) + NET_IP_ALIGN + extra_headroom);
	skb_reset_network_header(skb);

	// do not copy eth header from userspace
	data += sizeof(struct ethhdr);
	data_len -= sizeof(struct ethhdr);
	new_skb_data = skb_put(skb, data_len);
	skb->kct.next_hop = 0;
	skb->kct.next_hop6 = in6addr_any;

  	if ((ret = copy_from_user(new_skb_data, data, data_len))) {
		TRACE(LOG_WARN, "%s: copy_from_user(): segmentation fault.", __func__);
		kfree_skb(skb);
		return -EFAULT;
	}

	if (ver == 6) {
		if (route_info->ifindex == 0 && in6_addr_eq(route_info->next_hop6, in6addr_any)) {
			// policy routing not requested, but I need to set dst, so route to ip->daddr
			route_info->next_hop6 = ipv6_hdr(skb)->daddr;
		}
		
		if ((ret = pkt_policy_routing_ip6(skb, route_info, false))) {
			TRACE(LOG_WARN, "%s: pkt_policy_routing_ip6(): %d", __func__, ret);
			kfree_skb(skb);
			return -EINVAL;
		}
	}
	else {
		if (route_info->ifindex == 0 && route_info->next_hop == 0) {
			// policy routing not requested, but I need to set dst, so route to ip->daddr
			route_info->next_hop = ip_hdr(skb)->daddr;
		}
	
		if ((ret = pkt_policy_routing(skb, route_info, false))) {
			TRACE(LOG_WARN, "%s: pkt_policy_routing(): %d", __func__, ret);
			kfree_skb(skb);
			return -EINVAL;
		}
	}

	TRACE_DBG("%s: skb %p will be sent through '%s'.\n", __func__, skb, skb->dev->name);

	/* we don't want to see indicated packet again */
	skb->mark = PKT_SOCKET_TAG;
	
	if (ver == 6) {
		if ((ret = NF_HOOK_COND(PF_INET6, NF_INET_POST_ROUTING, skb, NULL, skb->dev, ip6_finish_output, 1))) {
			TRACE(LOG_WARN, "%s: ip6_finish_output(): %d", __func__, ret);
			return ret;
		}
	}
	else {
		if ((ret = NF_HOOK_COND(PF_INET, NF_INET_POST_ROUTING, skb, NULL, skb->dev, ip_finish_output, 1))) {
			TRACE(LOG_WARN, "%s: ip_finish_output(): %d", __func__, ret);
			return ret;
		}
	}

	return 0;
}

int32_t pkt_tx_rx_eth(unsigned char *data, unsigned int data_len, bool tx, struct kipf_route_info *route_info)
{
	struct sk_buff *skb = NULL;
	u_int8_t *dst = NULL;
	int ret = 0;

	if (!data) {
		TRACE(LOG_WARN, "%s: invalid parameter.", __func__);
		return -EINVAL;
	}

	if (data_len < sizeof(struct ethhdr) + sizeof(struct iphdr)) {
		TRACE(LOG_WARN, "%s: packet length is too short: %d", __func__, data_len);
		return -EINVAL;
	}

	if (!(skb = alloc_skb(data_len + NET_IP_ALIGN, GFP_ATOMIC))) {
		TRACE(LOG_WARN, "%s: alloc_skb(): can't allocate %d bytes of memory.", __func__, data_len);
		return -ENOMEM;
	}

	TRACE_DBG("%s: skb %p allocated.\n", __func__, skb);

	/* reset headers - specific header = (data - head) = 0 */
	skb_reset_mac_header(skb);
	skb_reset_network_header(skb);
	skb_reset_transport_header(skb);

	/* align skb to 4 bytes boundary - eth has 14 bytes (increase data and tail) */
	skb_reserve(skb, NET_IP_ALIGN /* 2 */);

	/* increase tail and len with data_len */
	dst = skb_put(skb, data_len);

	/* copy user data */
	if (copy_from_user(dst, (u_int8_t __user *)data, data_len)) {
		TRACE(LOG_WARN, "%s: copy_from_user(): segmentation fault.", __func__);
		kfree_skb(skb);
		return -EFAULT;
	}

	/* set ETH header */
	skb_set_mac_header(skb, NET_IP_ALIGN);

	/* set IP header */
	skb_set_network_header(skb, sizeof(struct ethhdr));

	/* set length of MAC header */
	skb->mac_len = sizeof(struct ethhdr);

	/* we don't want to see indicated packet again */
	skb->mark = PKT_SOCKET_TAG;

	/* find device */
	if (route_info->ifindex) {
		skb->dev = netdev_find_by_ifindex(route_info->ifindex);
	} else if (route_info->next_hop) {
		/* not implemented */
		skb->dev = NULL;
	} else {
		/* last chance */
		skb->dev = netdev_find_by_mac(((struct ethhdr*)dst)->h_dest, ETH_ALEN);
	}

	if (!IS_PTR_VALID(skb->dev)) {
		TRACE(LOG_WARN, "%s: can't find net device with ifindex %d, next_hop %u.%u.%u.%u",
			__func__, route_info->ifindex, IP_ADDRESS_STRING(route_info->next_hop));
		kfree_skb(skb);
		return -EINVAL;
	}

	/* loopback device? */
	if (skb->dev->flags & IFF_LOOPBACK) {
		TRACE(LOG_WARN, "%s: packet reception on loopback device '%s' isn't allowed.", __func__, skb->dev->name);
		kfree_skb(skb);
		return -EINVAL;
	}

	/* non ethernet or ppp device? */
	if (skb->dev->type != ARPHRD_ETHER && skb->dev->type != ARPHRD_PPP) {
		TRACE(LOG_WARN, "%s: packet reception on non eth or ppp device '%s' isn't allowed.", __func__, skb->dev->name);
		kfree_skb(skb);
		return -EINVAL;
	}

	if (tx) {
		skb->protocol = ((struct ethhdr*)skb_mac_header(skb))->h_proto;
		TRACE_DBG("%s: skb %p will be sent throught '%s'.\n", __func__, skb, skb->dev->name);
		if ((ret =  dev_queue_xmit(skb)) < 0) {
			TRACE(LOG_ERR, "%s: dev_queue_xmit(): %d", __func__, ret);
		}
	} else {
		skb->protocol = eth_type_trans(skb, skb->dev);
		TRACE_DBG("%s: skb %p will be received throught '%s'.\n", __func__, skb, skb->dev->name);
		if ((ret = netif_rx_ni(skb))) {
			TRACE(LOG_ERR, "%s: netif_rx_ni(): %d", __func__, ret);
		}
	}
	return ret;
}
